import os
from typing import List, Union, Optional

from prance import ResolvingParser

from caffoa.converter import parse_type, to_camelcase, TEMPLATE_FOLDER


class EndPoint:
    def __init__(self, name, path, operation, parameters: list, documentation: list, needs_content: bool):
        self.name = name
        self.parameters = parameters
        self.operation = operation
        self.path = path
        self.documentation = documentation
        self.needs_content = needs_content

    def __str__(self):
        return f"{self.operation} {self.path} -> {self.parameters}"


class Parameter:
    def __init__(self, name: str, type: str, desc: str):
        self.desc = desc
        self.type = type
        self.name = name


def create_function_files(endpoints: List[EndPoint], output_path: str, functions_name: str, namespace: str,
                          interface_target_folder: str, interface_name: str, interface_namespace: str,
                          boilerplate: Optional[str], imports: List[str]):
    if boilerplate is None:
        boilerplate = "{BASE}"

    with open(TEMPLATE_FOLDER + "/FunctionMethod.cs", "r", encoding="utf-8") as f:
        method_template = f.read()
    with open(TEMPLATE_FOLDER + "/FunctionTemplate.cs", "r", encoding="utf-8") as f:
        class_template = f.read()
    with open(TEMPLATE_FOLDER + "/InterfaceMethod.cs", "r", encoding="utf-8") as f:
        interface_method_template = f.read()
    with open(TEMPLATE_FOLDER + "/InterfaceTemplate.cs", "r", encoding="utf-8") as f:
        interface_template = f.read()
    methods = list()
    interface_methods = list()
    for ep in endpoints:
        params_with_names = list()
        params_with_names_for_interface = list()
        param_names = list()
        additioanl_error_infos = list()
        for param in ep.parameters:
            params_with_names.append(f"{param.type} {param.name}")
            params_with_names_for_interface.append(f"{param.type} {param.name}")
            param_names.append(param.name)
            additioanl_error_infos.append(f'debugInformation["p_{param.name}"] = {param.name}.ToString();\n\t\t\t\t')
        if ep.needs_content:
            param_names.append("req.Content")
            params_with_names_for_interface.append("HttpContent contentPayload")
        params_for_interface = ", ".join(params_with_names_for_interface)
        params_for_function = ", " + ", ".join(params_with_names) if len(params_with_names) > 0 else ""
        param_name_str = ", ".join(param_names)
        function_call = f"{ep.name}({param_name_str})"
        default_call = f"return await Service(req, log).{function_call};"
        invocation = boilerplate.replace("{BASE}",default_call).replace("{CALL}", function_call).replace("\n","\n\t\t\t\t").strip()
        methods.append(
            method_template.format(NAME=ep.name, OPERATION=ep.operation, PATH=ep.path, INVOCATION=invocation, PARAMS=params_for_function,
                                   PARAM_NAMES=param_name_str, ADDITIONAL_ERROR_INFOS="".join(additioanl_error_infos)))

        interface_methods.append(interface_method_template.format(NAME=ep.name, OPERATION=ep.operation, PATH=ep.path,
                                                                  PARAMS=params_for_interface,
                                                                  DOC="\n\t\t/// ".join(ep.documentation)))

    imports = map(lambda x: f"using {x};\n", imports)
    function_file_name = os.path.abspath(f"{output_path}/{functions_name}.generated.cs")
    print(f"Writing Function to {function_file_name}")
    with open(function_file_name, "w", encoding="utf-8") as f:
        f.write(class_template.format(METHODS="\n\n".join(methods), NAMESPACE=namespace, CLASSNAME=functions_name,
                                      INTERFACENAME=interface_name,
                                      IMPORTS="".join(imports)))
    interface_file_name = os.path.abspath(f"{interface_target_folder}/{interface_name}.generated.cs")

    print(f"Writing Interface to {interface_file_name}")
    with open(interface_file_name, "w", encoding="utf-8") as f:

        f.write(interface_template.format(METHODS="\n\n".join(interface_methods), NAMESPACE=interface_namespace,
                                          CLASSNAME=interface_name))


def parse_params(params: list) -> list:
    parameters = list()
    for param in params:
        if param['in'].lower() == "path":
            parameters.append(Parameter(param['name'], parse_type(param['schema']), param.get('description')))
    return parameters


def generate_functions(input_file: str, output_path: str, class_name: Union[str, None], namespace: str,
                       interface_target_folder: str, interface_name: str, interface_namespace: str,
                       boilerplate: Optional[str], imports: List[str]):
    if class_name is None:
        class_name = namespace
    parser = ResolvingParser(input_file, strict=False)
    endpoints = list()
    os.makedirs(output_path, exist_ok=True)
    os.makedirs(interface_target_folder, exist_ok=True)
    for path, options in parser.specification['paths'].items():
        path = str(path).strip('/')
        base_parameters = list()
        if "parameters" in options:
            base_parameters = parse_params(options["parameters"])
        for operation, config in options.items():
            operation = operation.lower()
            if operation in ["parameters"]:
                continue
            if "operationId" not in config:
                raise Warning(f"OperationId is missing for {path}: {operation}")
            operation_id = to_camelcase(config['operationId']) + "Async"
            documentation = config['description'].split("\n")
            for response, response_data in config["responses"].items():
                documentation.append(f"{response} -> {response_data['description']}")
            parameters = base_parameters.copy()
            if "parameters" in config:
                parameters.extend(parse_params(config["parameters"]))
            needs_content = "requestBody" in config
            endpoints.append(EndPoint(operation_id, path, operation, parameters, documentation, needs_content))

    create_function_files(endpoints, output_path, class_name, namespace,
                          interface_target_folder, interface_name, interface_namespace,
                          boilerplate, imports)



import argparse
import yaml

from caffoa.function import generate_functions
from caffoa.schema import generate_schemas


def execute():
    parser = argparse.ArgumentParser()
    parser.add_argument("--config", default="caffoa.yml", help="Path to config file (Default: caffoa.yml)")
    args = parser.parse_args()
    with open(args.config, "r", encoding="utf-8") as f:
        data = yaml.safe_load(f)
    if not "services" in data:
        raise Warning("no services in config file.")
    services = data["services"]
    if type(services) != list:
        raise Warning("services should be list")
    for id, config in enumerate(services):
        if not "apiPath" in config:
            raise Warning(f"apiPath is required for service #{id}")
        api = config["apiPath"]
        if "function" in config:
            function = config["function"]
            if not "name" in function or not "namespace" in function or not "targetFolder" in function:
                raise Warning(f"function needs children 'name', 'namespace' and 'targetFolder' in service #{id}")
            name = function['name']
            namespace = function["namespace"]
            target_folder = function['targetFolder']
            boilerplate = function.get('boilerplate')
            interface_name = function.get('interfaceName', f"I{name}Service")
            functions_name = function.get('functionsName', f"{name}Functions")
            interface_namespace = function.get('interfaceNamespace', namespace)
            interface_target_folder = function.get('interfaceTargetFolder', namespace)
            imports = function.get('imports', list())
            generate_functions(api, target_folder, functions_name, namespace, interface_target_folder, interface_name,
                               interface_namespace, boilerplate, imports)

        if "model" in config:
            model = config["model"]
            prefix = model.get("prefix", "")
            suffix = model.get("suffix", "")
            if not "namespace" in model or not "targetFolder" in model:
                raise Warning(f"model needs children 'namespace' and 'targetFolder' in service #{id}")
            excludes = list(model.get('excludes', list()))
            generate_schemas(api, model["targetFolder"], model['namespace'], prefix, suffix, excludes)



from typing import Tuple, Optional, List, Set
import re

from prance import BaseParser

from caffoa.converter import parse_type, to_camelcase


class MemberData:
    def __init__(self, name: str, type: str, is_date: bool, nullable: bool, default_value: Optional[str],
                 is_required: bool, description: Optional[str], enums: Optional[str]):
        self.name = name
        self.type = type
        self.is_date = is_date
        self.nullable = nullable
        self.default_value = default_value
        self.is_required = is_required
        self.description = description
        self.enums = enums


class ModelData:
    def __init__(self, rawname: str, name: str, parents: list, properties: List[MemberData], imports: Set[str],
                 description: Optional[str]):
        self.rawname = rawname
        self.name = name
        self.parents = parents
        self.properties = properties
        self.imports = imports
        self.description = description


class SchemaParser:
    def __init__(self, prefix: str, suffix: str, excludes: list):
        self.prefix = prefix
        self.suffix = suffix
        self.known_types = dict()
        self.excludes = excludes

    def parse(self, input_file: str) -> List[ModelData]:
        parser = BaseParser(input_file, strict=False)
        schemas = parser.specification["components"]["schemas"]
        for class_name, schema in schemas.items():
            if class_name in self.excludes:
                continue
            self.parse_simple(class_name, schema)

        objects = list()
        for class_name, schema in schemas.items():
            if self.class_name(class_name) in self.known_types:
                continue
            if class_name in self.excludes:
                continue
            objects.append(self.parse_objects(class_name, schema))
        return objects

    def parse_simple(self, class_name: str, schema: dict):
        if "type" in schema:
            typename = schema["type"]
            if typename in ["string", "integer", "number", "boolean"]:
                class_name = self.class_name(class_name)
                self.known_types[class_name] = schema

    def parse_objects(self, class_name: str, schema: dict, parents=None):
        if parents is None:
            parents = set()
        imports = set()
        members = list()
        if "allOf" in schema:
            return self.handle_all_of(class_name, schema, parents)

        if "properties" in schema:
            required_props = schema.get('required', list())
            for name, data in schema["properties"].items():

                # handle default type references
                if "$ref" in data:
                    typename = self.class_name_from_ref(data["$ref"])
                    if typename in self.known_types:
                        data = self.known_types[typename]

                enums = None
                isdate = False
                required = name in required_props
                nullable = "nullable" in data and data["nullable"]

                # handle nullable references that are constructed with anyOf
                if "anyOf" in data:
                    data = self.handle_any_of(data, name, class_name)

                description = None
                if "description" in data:
                    description = data["description"]

                if "$ref" in data:
                    typename, default_value = self.handle_ref(data, required)
                    nullable = not required

                elif "type" not in data:
                    print(data)
                    raise Warning(f"Cannot parse property without type for '{name}' in '{class_name}'")

                elif data["type"] == "array":
                    typename, default_value = self.handle_array(data, name, class_name, nullable)
                    imports.add("using System.Collections.Generic;\n")

                elif data["type"] == "object":
                    if "properties" in data:
                        raise Warning(
                            f"Cannot parse property trees: '{name}' child of '{class_name}' should be declared in own schema directly under 'components'")
                    if "additionalProperties" in data:
                        typename, default_value = self.handle_additional_properties(data, nullable)
                        imports.add("using System.Collections.Generic;\n")
                    else:
                        raise Warning(
                            f"Cannot nested object without additionalProperties: '{name}' child of '{class_name}'")

                else:
                    typename, default_value = self.handle_default_type(data, nullable)
                    enums = self.handle_enums(name, data)
                    if enums != None:
                        imports.add("using System.Collections.Immutable;\n")
                    isdate = self.is_date(data)

                members.append(
                    MemberData(name, typename, isdate, nullable, default_value, name in required_props, description,
                               enums))
        description = None
        if "description" in schema:
            description = schema["description"]

        return ModelData(class_name, self.class_name(class_name), parents, members, imports, description)

    def handle_any_of(self, data: dict, name: str, class_name: str):
        anyof = data["anyOf"]
        if len(anyof) != 1:
            raise Warning(f"anyOf with multiple children not supported for '{name}' in '{class_name}'")
        anyof = anyof[0]
        if "$ref" not in anyof:
            raise Warning(f"anyOf child other than $ref not supported for '{name}' in '{class_name}'")
        data["$ref"] = anyof["$ref"]
        typename = self.class_name_from_ref(data["$ref"])
        if typename in self.known_types:
            data = self.known_types[typename]
        return data

    def handle_ref(self, data: dict, required: bool) -> Tuple[str, str]:
        typename = self.class_name_from_ref(data["$ref"])

        if required:
            default_value = f"new {typename}()"
        else:
            default_value = "null"
        return typename, default_value

    def handle_array(self, data: dict, name: str, class_name: str, nullable: bool) -> Tuple[str, str]:
        items = data["items"]
        if "$ref" in items:
            typename = self.class_name_from_ref(items["$ref"])
        elif items["type"] not in ["string", "number", "integer", "boolean"]:
            raise Warning(
                f"Cannot parse array trees: '{name}' child of '{class_name}' should have it's array item declared in own schema directly under 'components'")
        else:
            typename = parse_type(items, nullable)
        typename = f"List<{typename}>"
        default_value = f"new {typename}()"
        return typename, default_value

    def handle_additional_properties(self, data: dict, nullable: bool) -> Tuple[str, str]:
        props = data['additionalProperties']
        if "$ref" in props:
            typename = self.class_name_from_ref(props["$ref"])
        else:
            typename = parse_type(props, nullable)
        typename = f"Dictionary<string, {typename}>"
        default_value = f"new {typename}()"
        return typename, default_value

    def handle_default_type(self, data: dict, nullable: bool) -> Tuple[str, str]:
        default_value = None
        if "default" in data:
            default_value = data["default"]
            if type(default_value) is str:
                default_value = f'"{default_value}"'
            elif type(default_value) is bool:
                default_value = "true" if default_value else "false"
            elif default_value is None:
                default_value = "null"
        elif nullable:
            default_value = "null"
        typename = parse_type(data, nullable and default_value == "null")
        return typename, default_value

    @staticmethod
    def is_date(schema: str):
        if "format" not in schema or "type" not in schema:
            return False
        return schema['type'] == "string" and schema['format'] == "date"

    def class_name_from_ref(self, param: str):
        name = param.split('/')[-1]
        return self.class_name(name)

    def class_name(self, name):
        return self.prefix + to_camelcase(name) + self.suffix

    def handle_all_of(self, class_name: str, schema: dict, parents: set):
        to_generate = None
        for element in schema["allOf"]:
            if "$ref" in element:
                parents.add(self.class_name_from_ref(element["$ref"]))
            if "type" in element or "properties" in element:
                if to_generate is not None:
                    raise Warning(
                        "allOf is implemented as inheritance; cannot have to children of allOf with direct implementation")
                to_generate = element
        if to_generate is None:
            print("WARNING: Creating class without content, no child of allOf is type object")
            to_generate = dict()
        return self.parse_objects(class_name, to_generate, parents)

    def handle_enums(self, name: str, data: dict):
        if "enum" not in data:
            return None
        enum = data["enum"]
        if len(enum) == 0:
            return None
        typename, _ = self.handle_default_type(data, False)
        if typename not in ["string", "int", "long", "double"]:
            return None

        names = list()
        lines = list()
        for value in enum:
            varname = to_camelcase(name) +to_camelcase(value) + "Value"
            varname = re.sub(r"[^A-Za-z0-9]+", '_', varname)
            names.append(varname)
            if typename == "string":
                value = f'"{value}"'
            line = f"public const {typename} {varname} = {value};"
            lines.append(line)


        listname = "AllowedValuesFor" + to_camelcase(name)
        values = ", ".join(names)
        line = f"public static readonly ImmutableArray<{typename}> {listname} = new ImmutableArray<{typename}>() {{ {values} }};"
        lines.append(line)
        return "\n\t\t".join(lines)



import os

from caffoa.converter import to_camelcase, TEMPLATE_FOLDER
from caffoa.schema_parser import SchemaParser, ModelData


def write_model(model: ModelData, output_path: str, namespace: str):
    with open(TEMPLATE_FOLDER + "/ModelTemplate.cs", "r", encoding="utf-8") as f:
        model_Template = f.read()
    with open(TEMPLATE_FOLDER + "/ModelPropertyTemplate.cs", "r", encoding="utf-8") as f:
        prop_Template = f.read()
    properties = list()
    imports = "".join(model.imports)
    if len(model.parents) > 0:
        parents = " : " + ", ".join(model.parents)
    else:
        parents = ""
    has_dates = False
    for property in model.properties:
        if property.is_date:
            extra = "[JsonConverter(typeof(CustomJsonDateConverter))]\n\t\t"
            has_dates = True
        else:
            extra = ""
        default_str = ""
        if property.default_value != None:
            default_str = f" = {property.default_value};"
        json_property_extra = ""
        if property.is_required and property.nullable:
            json_property_extra = ", Required = Required.AllowNull"
        elif property.is_required:
            json_property_extra = ", Required = Required.Always"
        elif not property.nullable:
            json_property_extra = ", Required = Required.DisallowNull"

        description = ""
        if property.description != None:
            prop_description = property.description.strip().replace("\n", "\n\t\t/// ")
            description = f"/// <summary>\n\t\t/// {prop_description}\n\t\t/// </summary>\n\t\t"

        enums = ""
        if property.enums != None:
            enums= f"{property.enums}\n\n\t\t"

        properties.append(
            prop_Template.format(TYPE=property.type, NAMELOWER=property.name, NAMEUPPER=to_camelcase(property.name),
                                 JSON_EXTRA=extra, DEFAULT=default_str, JSON_PROPERTY_EXTRA=json_property_extra,
                                 DESCRIPTION=description, ENUMS=enums))
    file_name = os.path.abspath(output_path + f"/{model.name}.generated.cs")
    print(f"Writing class {model.name} -> {file_name}")
    description = "/// AUTOGENERED BY caffoa ///\n\t"
    if model.description != None:
        mod_description = model.description.strip().replace("\n","\n\t/// ")
        description += f"/// <summary>\n\t/// {mod_description}\n\t/// </summary>\n\t"
    with open(file_name, "w", encoding="utf-8") as f:
        f.write(model_Template.format(NAMESPACE=namespace, NAME=model.name, RAWNAME=model.rawname, PROPERTIES="\n\n".join(properties),
                                      IMPORTS=imports, PARENTS=parents, DESCRIPTION=description))
    if has_dates:
        with open(TEMPLATE_FOLDER + "/DateSerializer.cs", "r", encoding="utf-8") as f:
            date_converter_template = f.read()

        file_name = os.path.abspath(output_path + f"/CustomJsonDateConverter.generated.cs")
        print(f"Writing CustomJsonDateConverter -> {file_name}")
        with open(file_name, "w", encoding="utf-8") as f:
            f.write(date_converter_template.format(NAMESPACE=namespace))


def generate_schemas(input_file: str, output_path: str, namespace: str, prefix: str, suffix: str, excludes: list):
    parser = SchemaParser(prefix, suffix, excludes)
    models = parser.parse(input_file)
    os.makedirs(output_path, exist_ok=True)
    for model in models:
        write_model(model, output_path, namespace)



import os

TEMPLATE_FOLDER = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data/templates")


def capitalize_first(data: str) -> str:
    if len(data) > 1:
        return data[0].upper() + data[1:]
    return data


def to_camelcase(s):
    s = s.replace("-", "_")
    return ''.join(capitalize_first(x) or '_' for x in s.split('_'))


def is_date(schema: dict):
    if "format" not in schema or "type" not in schema:
        return False
    return schema['type'] == "string" and schema['format'] == "date"


def parse_type(schema: dict, nullable: bool = False) -> str:
    format = None
    if "format" in schema:
        format = schema['format'].lower()
    type = schema['type'].lower()
    if nullable:
        suffix = "?"
    else:
        suffix = ""
    if type == "string" and format == "uuid":
        return f"System.Guid{suffix}"
    if type == "string" and format in ["date-time", "date"]:
        return f"System.DateTime{suffix}"
    if type == "integer" and format == "int64":
        return f"long{suffix}"
    if type == "integer":
        return f"int{suffix}"
    if type == "number":
        return f"double{suffix}"
    if type == "boolean":
        return f"bool{suffix}"
    if type == "string":
        return "string" # no nullable suffix for string
    return f"{type}{suffix}"
